package RPS::Mail::EventReporter;
use warnings;
use strict;

use Digest::HMAC_SHA1;
use IO::Socket::INET;

use Carp;

my %eventmap = (
	'greylisted'    => 1,
	'ungreylisted'  => 2,
	'auto-spam'     => 3,
	'hand-spam'     => 4,
	'auto-ham'      => 5,
	'hand-ham'      => 6,
	'valid-recip'   => 7,
	'invalid-recip' => 8,
	'virus'         => 9,
);

our $VERSION;
$VERSION = '3.01';

sub address_is_nonprivate
{
	my($addr) = @_;

	if ($addr =~ /:/) {
		# Unique Local Unicast
		return 0 if $addr =~ /^f[cd]..:/i;

		# Link Local Unicast
		return 0 if $addr =~ /^fe[89ab]..:/i;
		return 1;
	}

	my($a, $b, $c, $d) = split(/\./, $addr);
	if ($a == 10                           ||
	    $a == 127                          ||
	    ($a == 172 && $b >=16 && $b <= 31) ||
	    ($a = 192  && $b == 168) ||
	    ($a >= 224) ) {
		return 0;
	}
	return 1;
}

sub uncompress_ipv6
{
	my ($addr) = @_;

	return '0:0:0:0:0:0:0:0' if ($addr eq '::');

	return $addr unless $addr =~ /::/;

	# Can have at most *one* pair of colons
	return $addr if $addr =~ /::.*::/;

	$addr = "0$addr" if $addr =~ /^::/;
	$addr .= 0 if $addr =~ /::$/;

	# Count number of colons
	my $colons = ($addr =~ tr/:/:/);
	return $addr if ($colons >= 8);

	my $missing = ':' . ('0:' x (8 - $colons));
	$addr =~ s/::/$missing/;
	return $addr;
}

sub build_packet
{
	my($self) = @_;

	return unless $self->{username};

	# Pack data in binary format to save space
	# Size is currently:
	# 	 1 byte,  protocol version = 2
	# 	 1 byte,  username length
	# 	 n bytes, username
	# 	 8 bytes, random
	# 	 4 bytes, unix time
	# 	 SUBRECORDS
	# 	10 bytes, truncated password SHA1

	my $packet = pack('C C/A* CCCCCCCC N',

		# Protocol version, currently 1
		$self->{proto_version},

		# Username.  Length computed by C/A* pack() template
		$self->{username},

		# 8 pseudorandom bytes
		int(rand(256)), int(rand(256)), int(rand(256)), int(rand(256)),
		int(rand(256)), int(rand(256)), int(rand(256)), int(rand(256)),

		# UNIX time
		time(),
	    );

	# The actual events.  We sort by key so testing is sane!
	my @keys = sort { $a cmp $b } (keys(%{$self->{events}}));

	my $ip4 = '';
	my $ip6 = '';
	my $ip4rep = '';
	my $ip6rep = '';

	my $addr;

	foreach my $key (@keys) {
		my $event = $self->{events}->{$key};
		if ($event->{alen} == 16) {
			$addr = pack('nnnnnnnn', (map { hex($_) } (split(/:/, $event->{addr}))));
		} else {
			$addr = pack('CCCC', (split(/\./, $event->{addr})));
		}
		if ($event->{count} > 1) {
			$addr .= pack('CC', $event->{event}, $event->{count});
			if ($event->{alen} == 16) {
				$ip6rep .= $addr;
			} else {
				$ip4rep .= $addr;
			}
		} else {
			$addr .= pack('C', $event->{event});
			if ($event->{alen} == 16) {
				$ip6 .= $addr;
			} else {
				$ip4 .= $addr;
			}
		}
	}

	# If we have software name, include it, then version too if we have
	# that.
	if (defined($self->{software_name})) {
		$packet .= pack('Cn', 6, length($self->{software_name})) . $self->{software_name};
		if (defined($self->{software_version})) {
			$packet .= pack('Cn', 7, length($self->{software_version})) . $self->{software_version};
		}
	}

	$packet .= pack('Cn', 1, length($ip4)) .    $ip4    if ($ip4 ne '');
	$packet .= pack('Cn', 2, length($ip6)) .    $ip6    if ($ip6 ne '');
	$packet .= pack('Cn', 3, length($ip4rep)) . $ip4rep if ($ip4rep ne '');
	$packet .= pack('Cn', 4, length($ip6rep)) . $ip6rep if ($ip6rep ne '');

	# EOR record
	$packet .= pack('C', 0);

	my $ctx = Digest::HMAC_SHA1->new($self->{password});
	$ctx->add($packet);

	# Keep 80 bits of the SHA1 sum.
	return $packet . substr($ctx->digest, 0, 10);
}

sub report_ip_event
{
	my($self, $ip_addr, $event) = @_;

	# Map event name to a number
	confess("Invalid event '$event'") unless $eventmap{$event};
	$event = $eventmap{$event};

	# Do nothing if we have no user name
	return unless $self->{username};

	# Do nothing if it's a non-routable IP address
	return unless address_is_nonprivate($ip_addr);

	# Push it on the queue
	$self->_push_event($ip_addr, $event);
}

sub set_software_and_version
{
	my ($self, $software_name, $software_version) = @_;

	if (defined($software_name)) {
		if (length($software_name) > 63) {
			croak("set_software_and_version: Software name can be at most 63 characters long");
		} elsif ($software_name eq '') {
			croak("set_software_and_version: Software name must be at least 1 character long.");
		}
	}
	if (defined($software_version)) {
		if (length($software_version) > 31) {
			croak("set_software_and_version: Software version can be at most 31 characters long");
		} elsif ($software_version eq '') {
			croak("set_software_and_version: Software version must be at least 1 character long.");
		}
	}
	if (!defined($software_name)) {
		$self->{software_name} = undef;
		$self->{software_version} = undef;
	} else {
		$self->{software_name} = $software_name;
		$self->{software_version} = $software_version;
	}
}

sub _push_event
{
	my($self, $ip_addr, $eventno) = @_;

	my $alen;

	# Hack: If it contains a colon, it's IPv6
	if ($ip_addr =~ /:/) {
		$ip_addr = uncompress_ipv6($ip_addr);
		$alen = 16;
	} else {
		my @bytes = split(/\./, $ip_addr);
		return unless scalar(@bytes) == 4;

		foreach my $byte (@bytes) {
			return unless $byte =~ /^\d+$/;
			return if $byte > 255;
		}
		$alen = 4;
	}

	# Unique key for event
	my $key = "$ip_addr=$eventno";
	my $force_flush = 0;
	if (!exists($self->{events}->{$key})) {
		# A new nonrepeated event
		$self->{event_queue_current_bytes} += ($alen + 1);
		$self->{events}->{$key} = { addr => $ip_addr,
					    event => $eventno,
					    alen => $alen,
					    count => 1};
	} else {
		my $event = $self->{events}->{$key};
		if ($event->{count} == 1) {
			# Convert nonrepeated event to repeated event
			$self->{event_queue_current_bytes}++;
		}
		$event->{count}++;
		if ($event->{count} == 255) {
			# Force a flush if an event's count ever reaches 255!
			$force_flush = 1;
		}
	}
	if ($force_flush || $self->{event_queue_current_bytes} >= $self->{event_queue_max_bytes}) {
		$self->flush();
	}

}

sub new
{
	my ($class, $args) = @_;

	my $self = bless {
		report_host => undef,
		report_port => 6568,
		proto_version => 2,
		username => undef,
		password => undef,
		events => {},
		event_queue_max_bytes => 400,
		event_queue_current_bytes => 0,
		software_name => undef,
		software_version => undef,
	}, $class;
	$self->{username}    = delete($args->{user}) if exists($args->{user});
	$self->{password}    = delete($args->{password}) if exists($args->{password});
	$self->{report_host} = delete($args->{report_host}) if exists($args->{report_host});
	$self->{report_port} = delete($args->{report_port}) if exists($args->{report_port});

	croak("Must supply `user', `password' and `report_host' arguments to new")
	    unless defined($self->{username}) &&
	    defined($self->{password}) &&
	    defined($self->{report_host});
	croak('Unknown argument(s) to new: ' . join(', ', (map { "`$_'" } (sort(keys(%$args))))))
	    if (scalar(keys(%$args)));
	return $self;
}

sub flush
{
	my ($self) = @_;
	return unless $self->{event_queue_current_bytes} > 0;

	my $packet = $self->build_packet();

	# Don't need events any longer
	$self->{events} = {};
	$self->{event_queue_current_bytes} = 0;

	return unless $packet;

	my $sock = IO::Socket::INET->new(PeerAddr => $self->{report_host},
					 PeerPort => $self->{report_port},
					 Proto    => 'udp',
					 Timeout  => 5);
	return unless $sock;

	my $peer = $sock->peeraddr();
	$peer = unpack('C', $peer);

	# If the peer address is 127.x.x.x, it means Roaring Penguin
	# does NOT want reputation feedback, so do nothing
	if ($peer == 127) {
		$sock->close();
		return;
	}

	$sock->print($packet);
	$sock->close();
}

1;

__END__

=head1 NAME

RPS::Mail::EventReporter - Report IP address events to Roaring
Penguin's central reputation collector.

=head1 ABSTRACT

This class is used to report events to the central Roaring Penguin
reputation collector.

=head1 OVERVIEW

This module is used to report events to the central Roaring Penguin
reputation collector.  To use it in your MIMEDefang filter, you
should perform the following steps:

=head2 INITIALIZATION

First, initialize the reputation collector with your username and
password and the name of the host that will accept reports.  This
should be done once per filter process.  The call looks like this:

    use RPS::Mail::EventReporter;
    my $r = RPS::Mail::EventReporter->new({user => 'my_username',
                                           password => 'my_password',
                                           report_host => 'host'});
    $r->set_software_and_version('MIMEDefang/Myfilter', '2.68/1.02');

Substitute your username and password for 'username' and 'pass'.
Please contact Roaring Penguin Software Inc. to obtain credentials
to send reports as well as the host name to which reports should
be sent.

=head2 REPORTING

Next, report events.  If the variable $ip holds the IP address of the SMTP
client, then call the report_ip_event method as follows:

    $r->report_ip_event($ip, 'greylisted');
    $r->report_ip_event($ip, 'ungreylisted');
    $r->report_ip_event($ip, 'auto-spam');
    $r->report_ip_event($ip, 'auto-ham');
    $r->report_ip_event($ip, 'hand-spam');
    $r->report_ip_event($ip, 'hand-ham');
    $r->report_ip_event($ip, 'valid-recip');
    $r->report_ip_event($ip, 'invalid-recip');

The events must be reported under the following circumstances:

=over 4

=item * greylisted - The host was greylisted.

=item * ungreylisted - The host successfully passed greylisting.

=item * auto-spam - The host sent an e-mail automatically classified as spam.

=item * auto-ham - The host sent an e-mail automatically classified as non-spam.

=item * hand-spam - The host sent an e-mail that was classified I<by a human being> as spam.

=item * hand-ham - The host sent an e-mail that was classified I<by a human being> as non-spam.

=item * valid-recip - The host used a valid recipient in a RCPT command.

=item * invalid-recip - The host used an invalid recipient in a RCPT command.

=back

You I<must not> report events that your filter cannot reliably confirm to
be true.  For example, unless you have a mechanism to mark messages by
hand, you should I<not> report any hand-spam or hand-ham events.  If you are
not using greylisting, do not report any greylisted or ungreylisted events.
And if you are unable to validate recipients, do not report any
valid-recip or invalid-recip events.  Failure to adhere to these guidelines
could get your reporting account deleted.

Note that if you submit auto-ham and auto-spam reports, it should
I<not> be the case that an email is always reported as either
ham or spam.  You should have one threshold above which email
is reported as auto-spam.  You should have another threshold
below which email is reported as auto-ham.  There I<must> be
a reasonable gap between the levels for "uncertain" emails.
I<Do not> report an auto-ham or auto-spam event unless you are
very confident your filter has classified the email correctly.

=head2 CLEANUP

Just before your filtering process exits, it should flush any accumulated
data.  Do this as follows:

    $r->flush();

I<Do not> call flush() unless the filtering process is about to exit.

=head1 USE WITH MIMEDEFANG

The general flow of operation in a MIMEDefang filter looks something like this:

    # Outside any subroutine:
    use RPS::Mail::EventReporter;
    my $r;

    sub filter_initialize {
            $r = RPS::Mail::EventReporter->new({user => 'my_username',
                                                password => 'my_password',
                                                report_host => 'host'});
            $r->set_software_and_version('MIMEDefang/Myfilter', '2.68/1.02');
            # Any remaining filter_initialize work goes here
    }

    # Do all your reporting in filter_recipient, filter_begin/filter_end, etc.

    # Cleanup
    sub filter_cleanup {
            $r->flush();
            # Any other cleanup work goes here
    }

=head1 METHODS

The B<new> method is a class method that returns an object.  All other
methods are instance methods.

=head2 RPS::Mail::EventReporter->new ( $args )

Creates a reporting object.  $args is a hash with the following
possible keys:

=over 4

=item * user => I<username> - Specifies your username for reporting events.

=item * password => I<pass> - Specifies your password for reporting events.

=item * report_host => I<hostname> - Specifies the name of the host to which
reports are sent.

=item * report_port => I<port> - Specifies the UDP port to which to send
reports.  Defaults to 6568.

=back

The B<user>, B<password> and B<report_host> keys are required.
B<report_port> is optional.

=head2 set_software_and_version ( $software_name, $software_version)

Sets the software name and version to be reported back.  (This call is
optional; you do not need to report software name or version.)

If you supply only $software_name, then only a Software-Name subreport
is included in each packet.

=head2 report_ip_event ( $ipaddr, $event )

Reports the event $event for the given IP address to the central
Roaring Penguin reputation collector.

$event must be one of 'greylisted', 'ungreylisted', 'auto-spam',
'hand-spam', 'auto-ham', 'hand-ham', 'valid-recip' or 'invalid-recip'.
A report is not immediately sent; instead, events are queued and sent
when the queue reaches a certain size or flush() is called explicitly.

=head2 flush ( )

Flushes all queued events and (if necessary) sends a UDP packet
reporting the events.

=head2 build_packet ( )

Builds a UDP packet to report all queued events.  This is an internal
method and should not be called directly.

=head1 AUTHOR

   David F. Skoll, <dfs@roaringpenguin.com>

=head1 COPYRIGHT & LICENSE

MIT

Copyright (c) 2010 Roaring Penguin Software Inc.

 Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the
 "Software"), to deal in the Software without restriction, including
 without limitation the rights to use, copy, modify, merge, publish,
 distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to
 the following conditions:

 The above copyright notice and this permission notice shall be
 included in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
